import datetime
from django.utils.translation import ugettext_lazy as _
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .currency_pair import CurrencyPair


class ValueNotFoundException(Exception):
    """
    Exception raised when a value was not found for
    a currency pair on a given date

    Attributes:
        currency_pair -- CurrencyPair object
        date -- the date where the value was not found
    """

    def __init__(self, currency_pair: "CurrencyPair", date: datetime.date):
        self.currency_pair = currency_pair
        self.date = date
        self.message = _("Value for %(currency_pair)s on %(date)s was not found") % {
            "currency_pair": str(self.currency_pair),
            "date": str(self.date),
        }
        super().__init__(self.message)



# -*- encoding: utf-8 -*-
""" Models for the holidays application """

# standard library
import datetime
import json
from urllib.request import Request
from urllib.request import urlopen

# django
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .enums import Countries


class Holiday(models.Model):
    date = models.DateField(
        help_text=_("The date of the holiday"),
    )
    name = models.CharField(
        help_text=_("The name of the holiday"), max_length=100, verbose_name=_("name")
    )
    country_code = models.CharField(
        help_text=_("The code of the country this holiday belongs to."),
        max_length=2,
        verbose_name=_("country code"),
        choices=Countries.django_model_choices,
    )

    class Meta:
        ordering = ("date",)
        verbose_name = _('holiday')
        verbose_name_plural = _('holidays')

    def __str__(self):
        return f"{self.country_code}-{self.date}"

    @classmethod
    def update_holidays(cls, country_code):
        year = datetime.date.today().year

        request = Request(
            f'https://data.magnet.cl/api/v1/holidays/{country_code.lower()}/{year}/'
        )

        response = urlopen(request)
        data = json.loads(response.read())

        for holiday_data in data['objects']:
            date_string = holiday_data['date']
            date = datetime.datetime.strptime(date_string, '%Y-%m-%d').date()
            name = holiday_data['name']
            cls.objects.get_or_create(date=date, name=name, country_code=country_code)

        year += 1

        request = Request(
            f'https://data.magnet.cl/api/v1/holidays/{country_code.lower()}/{year}/'
        )

        response = urlopen(request)
        data = json.loads(response.read())

        for holiday_data in data['objects']:
            date_string = holiday_data['date']
            date = datetime.datetime.strptime(date_string, '%Y-%m-%d').date()
            name = holiday_data['name']
            cls.objects.get_or_create(
                date=date,
                name=name,
                country_code=country_code
            )



# -*- encoding: utf-8 -*-

# django
from django.db import models
from django.utils.translation import ugettext_lazy as _
from magnet_data.currencies.enums import CurrencyAcronyms


class CurrencyValue(models.Model):
    """
    This model stores the value that the `base_currency` currency has as
    `counter_currency` currency on the `date`
    """

    value = models.DecimalField(
        _("value"),
        decimal_places=6,
        max_digits=20,
    )
    date = models.DateField(
        verbose_name=_("currency date"),
        db_index=True,
    )
    base_currency = models.CharField(
        _("base currency"),
        max_length=5,
        help_text=_("The acronym of the base currency"),
        choices=CurrencyAcronyms.django_model_choices,
    )
    counter_currency = models.CharField(
        _("counter currency"),
        max_length=5,
        help_text=_("The acronym of the counter currency"),
        choices=CurrencyAcronyms.django_model_choices,
    )

    class Meta:
        verbose_name = _("currency value")
        verbose_name_plural = _("currency values")
        unique_together = (("base_currency", "counter_currency", "date"),)
        ordering = ("date",)

    def __str__(self) -> str:
        return f"{self.base_currency}/{self.counter_currency}-{self.date}-{self.value}"



# django
from django.utils.translation import ugettext_lazy as _


class Countries:
    AF = "AF"
    AX = "AX"
    AL = "AL"
    DZ = "DZ"
    AS = "AS"
    AD = "AD"
    AO = "AO"
    AI = "AI"
    AQ = "AQ"
    AG = "AG"
    AR = "AR"
    AM = "AM"
    AW = "AW"
    AU = "AU"
    AT = "AT"
    AZ = "AZ"
    BS = "BS"
    BH = "BH"
    BD = "BD"
    BB = "BB"
    BY = "BY"
    BE = "BE"
    BZ = "BZ"
    BJ = "BJ"
    BM = "BM"
    BT = "BT"
    BO = "BO"
    BQ = "BQ"
    BA = "BA"
    BW = "BW"
    BV = "BV"
    BR = "BR"
    IO = "IO"
    BN = "BN"
    BG = "BG"
    BF = "BF"
    BI = "BI"
    CV = "CV"
    KH = "KH"
    CM = "CM"
    CA = "CA"
    KY = "KY"
    CF = "CF"
    TD = "TD"
    CL = "CL"
    CN = "CN"
    CX = "CX"
    CC = "CC"
    CO = "CO"
    KM = "KM"
    CG = "CG"
    CD = "CD"
    CK = "CK"
    CR = "CR"
    CI = "CI"
    HR = "HR"
    CU = "CU"
    CW = "CW"
    CY = "CY"
    CZ = "CZ"
    DK = "DK"
    DJ = "DJ"
    DM = "DM"
    DO = "DO"
    EC = "EC"
    EG = "EG"
    SV = "SV"
    GQ = "GQ"
    ER = "ER"
    EE = "EE"
    SZ = "SZ"
    ET = "ET"
    FK = "FK"
    FO = "FO"
    FJ = "FJ"
    FI = "FI"
    FR = "FR"
    GF = "GF"
    PF = "PF"
    TF = "TF"
    GA = "GA"
    GM = "GM"
    GE = "GE"
    DE = "DE"
    GH = "GH"
    GI = "GI"
    GR = "GR"
    GL = "GL"
    GD = "GD"
    GP = "GP"
    GU = "GU"
    GT = "GT"
    GG = "GG"
    GN = "GN"
    GW = "GW"
    GY = "GY"
    HT = "HT"
    HM = "HM"
    VA = "VA"
    HN = "HN"
    HK = "HK"
    HU = "HU"
    IS = "IS"
    IN = "IN"
    ID = "ID"
    IR = "IR"
    IQ = "IQ"
    IE = "IE"
    IM = "IM"
    IL = "IL"
    IT = "IT"
    JM = "JM"
    JP = "JP"
    JE = "JE"
    JO = "JO"
    KZ = "KZ"
    KE = "KE"
    KI = "KI"
    KP = "KP"
    KR = "KR"
    KW = "KW"
    KG = "KG"
    LA = "LA"
    LV = "LV"
    LB = "LB"
    LS = "LS"
    LR = "LR"
    LY = "LY"
    LI = "LI"
    LT = "LT"
    LU = "LU"
    MO = "MO"
    MG = "MG"
    MW = "MW"
    MY = "MY"
    MV = "MV"
    ML = "ML"
    MT = "MT"
    MH = "MH"
    MQ = "MQ"
    MR = "MR"
    MU = "MU"
    YT = "YT"
    MX = "MX"
    FM = "FM"
    MD = "MD"
    MC = "MC"
    MN = "MN"
    ME = "ME"
    MS = "MS"
    MA = "MA"
    MZ = "MZ"
    MM = "MM"
    NA = "NA"
    NR = "NR"
    NP = "NP"
    NL = "NL"
    NC = "NC"
    NZ = "NZ"
    NI = "NI"
    NE = "NE"
    NG = "NG"
    NU = "NU"
    NF = "NF"
    MK = "MK"
    MP = "MP"
    NO = "NO"
    OM = "OM"
    PK = "PK"
    PW = "PW"
    PS = "PS"
    PA = "PA"
    PG = "PG"
    PY = "PY"
    PE = "PE"
    PH = "PH"
    PN = "PN"
    PL = "PL"
    PT = "PT"
    PR = "PR"
    QA = "QA"
    RE = "RE"
    RO = "RO"
    RU = "RU"
    RW = "RW"
    BL = "BL"
    SH = "SH"
    KN = "KN"
    LC = "LC"
    MF = "MF"
    PM = "PM"
    VC = "VC"
    WS = "WS"
    SM = "SM"
    ST = "ST"
    SA = "SA"
    SN = "SN"
    RS = "RS"
    SC = "SC"
    SL = "SL"
    SG = "SG"
    SX = "SX"
    SK = "SK"
    SI = "SI"
    SB = "SB"
    SO = "SO"
    ZA = "ZA"
    GS = "GS"
    SS = "SS"
    ES = "ES"
    LK = "LK"
    SD = "SD"
    SR = "SR"
    SJ = "SJ"
    SE = "SE"
    CH = "CH"
    SY = "SY"
    TW = "TW"
    TJ = "TJ"
    TZ = "TZ"
    TH = "TH"
    TL = "TL"
    TG = "TG"
    TK = "TK"
    TO = "TO"
    TT = "TT"
    TN = "TN"
    TR = "TR"
    TM = "TM"
    TC = "TC"
    TV = "TV"
    UG = "UG"
    UA = "UA"
    AE = "AE"
    GB = "GB"
    UM = "UM"
    US = "US"
    UY = "UY"
    UZ = "UZ"
    VU = "VU"
    VE = "VE"
    VN = "VN"
    VG = "VG"
    VI = "VI"
    WF = "WF"
    EH = "EH"
    YE = "YE"
    ZM = "ZM"
    ZW = "ZW"

    django_model_choices = (
        (AF, _("Afghanistan")),
        (AX, _("Åland Islands")),
        (AL, _("Albania")),
        (DZ, _("Algeria")),
        (AS, _("American Samoa")),
        (AD, _("Andorra")),
        (AO, _("Angola")),
        (AI, _("Anguilla")),
        (AQ, _("Antarctica")),
        (AG, _("Antigua and Barbuda")),
        (AR, _("Argentina")),
        (AM, _("Armenia")),
        (AW, _("Aruba")),
        (AU, _("Australia")),
        (AT, _("Austria")),
        (AZ, _("Azerbaijan")),
        (BS, _("Bahamas")),
        (BH, _("Bahrain")),
        (BD, _("Bangladesh")),
        (BB, _("Barbados")),
        (BY, _("Belarus")),
        (BE, _("Belgium")),
        (BZ, _("Belize")),
        (BJ, _("Benin")),
        (BM, _("Bermuda")),
        (BT, _("Bhutan")),
        (BO, _("Bolivia (Plurinational State of)")),
        (BQ, _("Bonaire, Sint Eustatius and Saba")),
        (BA, _("Bosnia and Herzegovina")),
        (BW, _("Botswana")),
        (BV, _("Bouvet Island")),
        (BR, _("Brazil")),
        (IO, _("British Indian Ocean Territory")),
        (BN, _("Brunei Darussalam")),
        (BG, _("Bulgaria")),
        (BF, _("Burkina Faso")),
        (BI, _("Burundi")),
        (CV, _("Cabo Verde")),
        (KH, _("Cambodia")),
        (CM, _("Cameroon")),
        (CA, _("Canada")),
        (KY, _("Cayman Islands")),
        (CF, _("Central African Republic")),
        (TD, _("Chad")),
        (CL, _("Chile")),
        (CN, _("China")),
        (CX, _("Christmas Island")),
        (CC, _("Cocos (Keeling) Islands")),
        (CO, _("Colombia")),
        (KM, _("Comoros")),
        (CG, _("Congo")),
        (CD, _("Congo (the Democratic Republic of the)")),
        (CK, _("Cook Islands")),
        (CR, _("Costa Rica")),
        (CI, _("Côte d'Ivoire")),
        (HR, _("Croatia")),
        (CU, _("Cuba")),
        (CW, _("Curaçao")),
        (CY, _("Cyprus")),
        (CZ, _("Czechia")),
        (DK, _("Denmark")),
        (DJ, _("Djibouti")),
        (DM, _("Dominica")),
        (DO, _("Dominican Republic")),
        (EC, _("Ecuador")),
        (EG, _("Egypt")),
        (SV, _("El Salvador")),
        (GQ, _("Equatorial Guinea")),
        (ER, _("Eritrea")),
        (EE, _("Estonia")),
        (SZ, _("Eswatini")),
        (ET, _("Ethiopia")),
        (FK, _("Falkland Islands (Malvinas)")),
        (FO, _("Faroe Islands")),
        (FJ, _("Fiji")),
        (FI, _("Finland")),
        (FR, _("France")),
        (GF, _("French Guiana")),
        (PF, _("French Polynesia")),
        (TF, _("French Southern Territories")),
        (GA, _("Gabon")),
        (GM, _("Gambia")),
        (GE, _("Georgia")),
        (DE, _("Germany")),
        (GH, _("Ghana")),
        (GI, _("Gibraltar")),
        (GR, _("Greece")),
        (GL, _("Greenland")),
        (GD, _("Grenada")),
        (GP, _("Guadeloupe")),
        (GU, _("Guam")),
        (GT, _("Guatemala")),
        (GG, _("Guernsey")),
        (GN, _("Guinea")),
        (GW, _("Guinea-Bissau")),
        (GY, _("Guyana")),
        (HT, _("Haiti")),
        (HM, _("Heard Island and McDonald Islands")),
        (VA, _("Holy See")),
        (HN, _("Honduras")),
        (HK, _("Hong Kong")),
        (HU, _("Hungary")),
        (IS, _("Iceland")),
        (IN, _("India")),
        (ID, _("Indonesia")),
        (IR, _("Iran (Islamic Republic of)")),
        (IQ, _("Iraq")),
        (IE, _("Ireland")),
        (IM, _("Isle of Man")),
        (IL, _("Israel")),
        (IT, _("Italy")),
        (JM, _("Jamaica")),
        (JP, _("Japan")),
        (JE, _("Jersey")),
        (JO, _("Jordan")),
        (KZ, _("Kazakhstan")),
        (KE, _("Kenya")),
        (KI, _("Kiribati")),
        (KP, _("Korea (the Democratic People's Republic of)")),
        (KR, _("Korea (the Republic of)")),
        (KW, _("Kuwait")),
        (KG, _("Kyrgyzstan")),
        (LA, _("Lao People's Democratic Republic")),
        (LV, _("Latvia")),
        (LB, _("Lebanon")),
        (LS, _("Lesotho")),
        (LR, _("Liberia")),
        (LY, _("Libya")),
        (LI, _("Liechtenstein")),
        (LT, _("Lithuania")),
        (LU, _("Luxembourg")),
        (MO, _("Macao")),
        (MG, _("Madagascar")),
        (MW, _("Malawi")),
        (MY, _("Malaysia")),
        (MV, _("Maldives")),
        (ML, _("Mali")),
        (MT, _("Malta")),
        (MH, _("Marshall Islands")),
        (MQ, _("Martinique")),
        (MR, _("Mauritania")),
        (MU, _("Mauritius")),
        (YT, _("Mayotte")),
        (MX, _("Mexico")),
        (FM, _("Micronesia (Federated States of)")),
        (MD, _("Moldova (the Republic of)")),
        (MC, _("Monaco")),
        (MN, _("Mongolia")),
        (ME, _("Montenegro")),
        (MS, _("Montserrat")),
        (MA, _("Morocco")),
        (MZ, _("Mozambique")),
        (MM, _("Myanmar")),
        (NA, _("Namibia")),
        (NR, _("Nauru")),
        (NP, _("Nepal")),
        (NL, _("Netherlands")),
        (NC, _("New Caledonia")),
        (NZ, _("New Zealand")),
        (NI, _("Nicaragua")),
        (NE, _("Niger")),
        (NG, _("Nigeria")),
        (NU, _("Niue")),
        (NF, _("Norfolk Island")),
        (MK, _("North Macedonia")),
        (MP, _("Northern Mariana Islands")),
        (NO, _("Norway")),
        (OM, _("Oman")),
        (PK, _("Pakistan")),
        (PW, _("Palau")),
        (PS, _("Palestine, State of")),
        (PA, _("Panama")),
        (PG, _("Papua New Guinea")),
        (PY, _("Paraguay")),
        (PE, _("Peru")),
        (PH, _("Philippines")),
        (PN, _("Pitcairn")),
        (PL, _("Poland")),
        (PT, _("Portugal")),
        (PR, _("Puerto Rico")),
        (QA, _("Qatar")),
        (RE, _("Réunion")),
        (RO, _("Romania")),
        (RU, _("Russian Federation")),
        (RW, _("Rwanda")),
        (BL, _("Saint Barthélemy")),
        (SH, _("Saint Helena, Ascension and Tristan da Cunha")),
        (KN, _("Saint Kitts and Nevis")),
        (LC, _("Saint Lucia")),
        (MF, _("Saint Martin (French part)")),
        (PM, _("Saint Pierre and Miquelon")),
        (VC, _("Saint Vincent and the Grenadines")),
        (WS, _("Samoa")),
        (SM, _("San Marino")),
        (ST, _("Sao Tome and Principe")),
        (SA, _("Saudi Arabia")),
        (SN, _("Senegal")),
        (RS, _("Serbia")),
        (SC, _("Seychelles")),
        (SL, _("Sierra Leone")),
        (SG, _("Singapore")),
        (SX, _("Sint Maarten (Dutch part)")),
        (SK, _("Slovakia")),
        (SI, _("Slovenia")),
        (SB, _("Solomon Islands")),
        (SO, _("Somalia")),
        (ZA, _("South Africa")),
        (GS, _("South Georgia and the South Sandwich Islands")),
        (SS, _("South Sudan")),
        (ES, _("Spain")),
        (LK, _("Sri Lanka")),
        (SD, _("Sudan")),
        (SR, _("Suriname")),
        (SJ, _("Svalbard and Jan Mayen")),
        (SE, _("Sweden")),
        (CH, _("Switzerland")),
        (SY, _("Syrian Arab Republic")),
        (TW, _("Taiwan (Province of China)")),
        (TJ, _("Tajikistan")),
        (TZ, _("Tanzania, the United Republic of")),
        (TH, _("Thailand")),
        (TL, _("Timor-Leste")),
        (TG, _("Togo")),
        (TK, _("Tokelau")),
        (TO, _("Tonga")),
        (TT, _("Trinidad and Tobago")),
        (TN, _("Tunisia")),
        (TR, _("Türkiye")),
        (TM, _("Turkmenistan")),
        (TC, _("Turks and Caicos Islands")),
        (TV, _("Tuvalu")),
        (UG, _("Uganda")),
        (UA, _("Ukraine")),
        (AE, _("United Arab Emirates")),
        (GB, _("United Kingdom of Great Britain and Northern Ireland")),
        (UM, _("United States Minor Outlying Islands")),
        (US, _("United States of America")),
        (UY, _("Uruguay")),
        (UZ, _("Uzbekistan")),
        (VU, _("Vanuatu")),
        (VE, _("Venezuela (Bolivarian Republic of)")),
        (VN, _("Viet Nam")),
        (VG, _("Virgin Islands (British)")),
        (VI, _("Virgin Islands (U.S.)")),
        (WF, _("Wallis and Futuna")),
        (EH, _("Western Sahara")),
        (YE, _("Yemen")),
        (ZM, _("Zambia")),
        (ZW, _("Zimbabwe")),
    )



from django.apps import apps
from magnet_data.currencies.currency_pair import CurrencyPair
from magnet_data.currencies.enums import CurrencyAcronyms
from magnet_data.holidays.enums import Countries
from magnet_data import utils
import datetime


class Currencies(CurrencyAcronyms):
    @staticmethod
    def get_pair(base_currency: str, counter_currency: str) -> CurrencyPair:
        """
        Returns a CurrencyPair object
        """
        return CurrencyPair(
            base_currency=base_currency,
            counter_currency=counter_currency
        )


class Holidays(Countries):
    def __init__(self):
        self.last_updated = None
        self.cls = apps.get_model(
            app_label='magnet_data',
            model_name='Holiday'
        )

    def update(self, country_code: str):
        """
        Update values stored in the database with what the api returned
        """
        now = datetime.datetime.now()
        threshold = now - datetime.timedelta(1)
        if self.last_updated is None or self.last_updated < threshold:
            self.last_updated = now
            self.cls.update_holidays(country_code=country_code.upper())

    def is_workday(self, date, country_code: str):
        """
        Returns True if the given date is not Saturday, Sunday, or
        Holiday
        Keyword arguments:
            from_date -- date to check
            country-code -- ISO 3166 country code
        """
        self.update(country_code)
        if date.weekday() == 5:  # Saturday
            return False

        if date.weekday() == 6:  # Sunday
            return False

        return not self.cls.objects.filter(
            date=date,
            country_code=country_code.upper(),
        ).exists()

    def get_next_working_day(self,
                             country_code: str,
                             working_days: int = 1,
                             from_date: datetime.date = None,
                             step: int = 1):
        """
        Returns the next date that is a working day.
        Keyword arguments:
            country-code -- ISO 3166 country code
            working_days -- number of working days to count (default 1)
            from_date -- date to start counting from (default today)
            step -- the amount by which the index increases. (default 1)
        """
        self.update(country_code)

        if from_date is None:
            from_date = utils.today()

        final_date = from_date

        while working_days > 0:
            final_date += datetime.timedelta(days=step)

            if self.is_workday(date=final_date, country_code=country_code):
                working_days -= 1

        return final_date

    def get_holidays_count_during_weekdays(self,
                                           country_code: str,
                                           start_date: datetime.date,
                                           end_date: datetime.date):
        """
        Returns the number of holidays between to days, not considering
        saturdays and sundays
        Keyword arguments:
            country-code -- de ISO 3166 country code
            start_date -- date to start counting from
            end_date -- date where to stop counting
        """
        self.update(country_code)
        days = 0

        holidays_dates = self.cls.objects.filter(
            date__range=[start_date, end_date],
            country_code=country_code.upper(),
        ).values_list('date', flat=True)

        for date in holidays_dates:
            if date.weekday() not in (5, 6):  # not in sunday or saturday
                days += 1

        return days


class MagnetDataClient:
    def __init__(self) -> None:
        super().__init__()
        self.currencies = Currencies()
        self.holidays = Holidays()



# standard library
from urllib.request import Request
from urllib.request import urlopen
import datetime
import json

# django
from django.apps import apps

# magnet data
from magnet_data.currencies.urls import API_URL


def get_url(year: int, month: int, base_currency: str, counter_currency: str) -> str:
    """
    Obtain the api url to get all the values of {base_currency} as {counter_currency}
    on month {month} of the year {year}
    """
    base_currency = base_currency.lower()
    counter_currency = counter_currency.lower()

    if month < 10:
        month = "0{}".format(month)

    return f"{API_URL}{base_currency}/{counter_currency}/{year}/{month}/"


def update_values(year: int, month: int, base_currency: str,
                  counter_currency: str) -> None:
    """
    Obtain all values of the {month}-{year} month from {base_currency} to
    {counter_currency}
    """
    url = get_url(year, month, base_currency, counter_currency)

    request = Request(url)

    response = urlopen(request)
    data = json.loads(response.read())

    for values_data in data["objects"]:
        date_string = values_data["date"]
        date = datetime.datetime.strptime(date_string, "%Y-%m-%d").date()

        CurrencyValue = apps.get_model(
            app_label='magnet_data',
            model_name='CurrencyValue'
        )

        CurrencyValue.objects.update_or_create(
            date=date,
            base_currency=base_currency,
            counter_currency=counter_currency,
            defaults={
                "value": values_data["value"],
            },
        )



import datetime
from django.utils import timezone


def today() -> datetime.date:
    """
    This method obtains today's date in local time
    """
    return timezone.localtime(timezone.now()).date()



# django
from django.contrib import admin

from holidays.models import Holiday


class HolidayAdmin(admin.ModelAdmin):
    list_display = ("name", "date", "country_code")


admin.site.register(Holiday, HolidayAdmin)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magnet_data',
 'magnet_data.currencies',
 'magnet_data.holidays',
 'magnet_data.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django>=2.2']

setup_kwargs = {
    'name': 'django-magnet-data',
    'version': '1.1.3',
    'description': 'An API client for data.magnet.cl',
    'long_description': None,
    'author': 'Ignacio Munizaga',
    'author_email': 'muni@magnet.cl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/magnet-cl/django-magnet-data',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# Generated by Django 3.2.14 on 2022-07-06 02:25

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CurrencyValue",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "value",
                    models.DecimalField(
                        decimal_places=6, max_digits=20, verbose_name="value"
                    ),
                ),
                ("date", models.DateField(db_index=True, verbose_name="currency date")),
                (
                    "base_currency",
                    models.CharField(
                        choices=[
                            ("CLP", "CLP"),
                            ("CLF", "CLF"),
                            ("USD", "USD"),
                            ("EUR", "EUR"),
                        ],
                        help_text="The acronym of the base currency",
                        max_length=5,
                        verbose_name="base currency",
                    ),
                ),
                (
                    "counter_currency",
                    models.CharField(
                        choices=[
                            ("CLP", "CLP"),
                            ("CLF", "CLF"),
                            ("USD", "USD"),
                            ("EUR", "EUR"),
                        ],
                        help_text="The acronym of the counter currency",
                        max_length=5,
                        verbose_name="counter currency",
                    ),
                ),
            ],
            options={
                "verbose_name": "currency value",
                "verbose_name_plural": "currency values",
                "ordering": ("date",),
                "unique_together": {("base_currency", "counter_currency", "date")},
            },
        ),
    ]



# standard library
from decimal import Decimal
import datetime

# django
from django.core.cache import cache
from django.apps import apps

# magnet data
from magnet_data import utils
from magnet_data.currencies.enums import CurrencyAcronyms
from magnet_data.currencies.exceptions import ValueNotFoundException
from magnet_data.currencies.client import update_values


class CurrencyPair:
    """
    A currency pair is the dyadic quotation of the relative value of a currency unit
    against the unit of another currency in the foreign exchange market. The currency
    that is used as the reference is called the counter currency, and the currency that
    is quoted in relation is called the base currency or transaction currency.

    https://en.wikipedia.org/wiki/Currency_pair

    Currently only handles pairs in which the counter currency is CLP
    """
    inverse_value = False

    def __init__(self, base_currency: str, counter_currency: str) -> None:
        super().__init__()
        if base_currency not in dict(CurrencyAcronyms.django_model_choices):
            raise ValueError(f"base_currency {base_currency} is not a valid choice")

        if counter_currency != CurrencyAcronyms.CLP:
            if base_currency != CurrencyAcronyms.CLP:
                raise NotImplementedError
            counter_currency, base_currency = base_currency, counter_currency
            self.inverse_value = True

        if counter_currency not in dict(CurrencyAcronyms.django_model_choices):
            raise ValueError(
                f"counter_currency {counter_currency} is not a valid choice"
            )

        self.base_currency = base_currency
        self.counter_currency = counter_currency

    def __str__(self) -> str:
        return f"{self.base_currency}/{self.counter_currency}"

    def cast_value(self, value):
        decimal_value = Decimal(value)
        if self.inverse_value:
            return 1 / decimal_value
        return decimal_value

    def last_knowable_date(self) -> datetime.date:
        today = utils.today()

        # only CLF values are known beforehand
        if self.base_currency != CurrencyAcronyms.CLF:
            return today

        if today.day < 10:
            return datetime.date(today.year, today.month, 9)
        elif today.month < 12:
            return datetime.date(today.year, today.month + 1, 9)

        return datetime.date(today.year + 1, 1, 9)

    def is_conversion_possible(self, date: datetime.date) -> bool:
        """
        Returns if the conversion is possible. Currently all future requests are being
        denied except for CLF conversions, since it's published beforehand
        """
        return self.last_knowable_date() >= date

    def on_date(self, date: datetime.date) -> Decimal:
        """returns the value for a given date"""
        if self.base_currency == self.counter_currency:
            return self.cast_value(1)

        if not self.is_conversion_possible(date):
            raise ValueNotFoundException(self, date)

        CurrencyValue = apps.get_model(
            app_label='magnet_data',
            model_name='CurrencyValue'
        )

        cache_key = f"md-{self.base_currency}/{self.counter_currency}/{date}"

        value = cache.get(cache_key)
        if value:
            return self.cast_value(value)

        queryset = CurrencyValue.objects.filter(
            base_currency=self.base_currency,
            counter_currency=self.counter_currency,
            date=date,
        )

        try:
            currency_value = queryset.get()
        except CurrencyValue.DoesNotExist:
            update_values(
                date.year, date.month, self.base_currency, self.counter_currency
            )

            try:
                currency_value = queryset.get()
            except CurrencyValue.DoesNotExist:
                raise ValueNotFoundException(self, date)

        cache.set(cache_key, currency_value.value)
        return self.cast_value(currency_value.value)

    def now(self) -> Decimal:
        """
        Return the current value of base_currency as counter_currency
        """
        return self.on_date(utils.today())

    def latest(self) -> Decimal:
        """
        Return the latest known value of the currency.
        If a currency has predefined values (like CLF) this will return future values.
        For everything else, it will return the same as self.now()
        """
        return self.on_date(self.last_knowable_date())

    def on_month(self, year: int, month: int) -> Decimal:
        raise NotImplementedError(f"{self.__class__.__name__}.on_month")



# django
from django.utils.translation import ugettext_lazy as _


class CurrencyAcronyms:
    CLF = "CLF"
    USD = "USD"
    EUR = "EUR"
    CLP = "CLP"

    django_model_choices = (
        (CLP, _("CLP")),
        (CLF, _("CLF")),
        (USD, _("USD")),
        (EUR, _("EUR")),
    )
